from typing import Protocol
from urllib.parse import urlparse

from ingestr.src.destinations import (
    BigQueryDestination,
    DatabricksDestination,
    DuckDBDestination,
    MsSQLDestination,
    PostgresDestination,
    RedshiftDestination,
    SnowflakeDestination,
)
from ingestr.src.sources import SqlSource

SQL_SOURCE_SCHEMES = [
    "bigquery",
    "duckdb",
    "mssql",
    "mysql",
    "postgres",
    "postgresql",
    "redshift",
    "snowflake",
    "sqlite",
]


class SourceProtocol(Protocol):
    def dlt_source(self, uri: str, table: str, **kwargs):
        pass


class DestinationProtocol(Protocol):
    def dlt_dest(self, uri: str, **kwargs):
        pass

    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        pass


class SourceDestinationFactory:
    source_scheme: str
    destination_scheme: str

    def __init__(self, source_uri: str, destination_uri: str):
        self.source_uri = source_uri
        source_fields = urlparse(source_uri)
        self.source_scheme = source_fields.scheme

        self.destination_uri = destination_uri
        dest_fields = urlparse(destination_uri)
        self.destination_scheme = dest_fields.scheme

    def get_source(self) -> SourceProtocol:
        if self.source_scheme in SQL_SOURCE_SCHEMES:
            return SqlSource()
        else:
            raise ValueError(f"Unsupported source scheme: {self.source_scheme}")

    def get_destination(self) -> DestinationProtocol:
        match: dict[str, DestinationProtocol] = {
            "bigquery": BigQueryDestination(),
            "databricks": DatabricksDestination(),
            "duckdb": DuckDBDestination(),
            "mssql": MsSQLDestination(),
            "postgres": PostgresDestination(),
            "postgresql": PostgresDestination(),
            "redshift": RedshiftDestination(),
            "snowflake": SnowflakeDestination(),
        }

        if self.destination_scheme in match:
            return match[self.destination_scheme]
        else:
            raise ValueError(
                f"Unsupported destination scheme: {self.destination_scheme}"
            )



from typing import Any, Optional

from dlt.common.schema.typing import TColumnSchema, TTableSchemaColumns
from sqlalchemy import Column, Table
from sqlalchemy.sql import sqltypes


def sqla_col_to_column_schema(sql_col: Column[Any]) -> Optional[TColumnSchema]:
    """Infer dlt schema column type from an sqlalchemy type.

    Precision and scale is inferred from that types that support it,
    such as numeric, varchar, int, bigint
    """
    sql_t = sql_col.type
    col = None

    if isinstance(sql_t, sqltypes.BigInteger):
        col = dict(name=sql_col.name, data_type="bigint", precision=64)
    elif isinstance(sql_t, sqltypes.SmallInteger):
        col = dict(name=sql_col.name, data_type="bigint", precision=16)
    elif isinstance(sql_t, sqltypes.Integer):
        col = dict(name=sql_col.name, data_type="bigint", precision=32)
    elif isinstance(sql_t, sqltypes.Numeric) and not isinstance(sql_t, sqltypes.Float):
        col = dict(
            name=sql_col.name,
            data_type="decimal",
            precision=sql_t.precision,
            scale=sql_t.scale,
        )
    elif isinstance(sql_t, sqltypes.String):
        col = dict(name=sql_col.name, data_type="text", precision=sql_t.length)
    elif isinstance(sql_t, sqltypes._Binary):
        col = dict(name=sql_col.name, data_type="binary", precision=sql_t.length)
    elif isinstance(sql_t, sqltypes.DateTime):
        col = dict(name=sql_col.name, data_type="timestamp")
    elif isinstance(sql_t, sqltypes.Date):
        col = dict(name=sql_col.name, data_type="date")
    elif isinstance(sql_t, sqltypes.Time):
        col = dict(name=sql_col.name, data_type="time")
    if col:
        return {key: value for key, value in col.items() if value is not None}  # type: ignore[return-value]
    return None


def table_to_columns(table: Table) -> TTableSchemaColumns:
    """Convert an sqlalchemy table to a dlt table schema.

    Only columns types supporting precision/scale are included in result.
    """
    return {
        col["name"]: col  # type: ignore
        for col in (sqla_col_to_column_schema(c) for c in table.columns)
        if col is not None
    }



import json
from urllib.parse import parse_qs, urlparse

import dlt


class GenericSqlDestination:
    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        table_fields = table.split(".")
        if len(table_fields) != 2:
            raise ValueError("Table name must be in the format <schema>.<table>")

        res = {
            "dataset_name": table_fields[-2],
            "table_name": table_fields[-1],
        }

        return res


class BigQueryDestination:
    def dlt_dest(self, uri: str, **kwargs):
        source_fields = urlparse(uri)
        source_params = parse_qs(source_fields.query)

        cred_path = source_params.get("credentials_path")
        if not cred_path:
            raise ValueError("Credentials path is required")

        location = None
        if source_params.get("location"):
            loc_params = source_params.get("location", [])
            if len(loc_params) > 1:
                raise ValueError("Only one location is allowed")
            location = loc_params[0]

        with open(cred_path[0], "r") as f:
            credentials = json.load(f)

        return dlt.destinations.bigquery(
            credentials=credentials, location=location, **kwargs
        )

    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        table_fields = table.split(".")
        if len(table_fields) != 2 and len(table_fields) != 3:
            raise ValueError(
                "Table name must be in the format <dataset>.<table> or <project>.<dataset>.<table>"
            )

        res = {
            "dataset_name": table_fields[-2],
            "table_name": table_fields[-1],
        }

        return res


class PostgresDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.postgres(credentials=uri, **kwargs)


class SnowflakeDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.snowflake(credentials=uri, **kwargs)


class RedshiftDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.redshift(credentials=uri, **kwargs)


class DuckDBDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.duckdb(credentials=uri, **kwargs)


class MsSQLDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.mssql(credentials=uri, **kwargs)


class DatabricksDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.databricks(credentials=uri, **kwargs)
